﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet
{
    /// <inheritdoc/>
    /// <typeparam name="TCodon">The type of the codons (fundamental building blocks) that compose the
    /// chromosome.</typeparam>
    public class Chromosome<TCodon> : Chromosome
    {
        public Chromosome(int length)
            : base(length)
        { 
        }

        public Chromosome()
            : base()
        {
        }

        public new TCodon this[int index]
        {
            get { return (TCodon)this.Codons[index]; }
            set { this.Codons[index] = value; }
        }

        public new virtual Chromosome<TCodon> Clone(bool copyIsOperable)
        {
            var clone = new Chromosome<TCodon>();
            CloneBase(clone, copyIsOperable);
            return clone;
        }
    }

    /// <summary>
    /// Represents the genetic information used to encode an individual member of a <see cref="T:Population" />.
    /// </summary>
    /// <remarks>
    /// <para>
    /// A chromosome (or genome) specifies the set of parameters that define a possible solution to a problem
    /// by encoding these parameters in a fixed-length array of objects. Before being evaluated, chromosomes 
    /// are usually transformed into a more manageable form.
    /// </para>
    /// </remarks>
    [System.Diagnostics.DebuggerDisplay("Fitness = {Fitness:#0}")]
    public class Chromosome
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="T:Chromosome"/> class, specifying its length.
        /// </summary>
        /// <param name="length">The length of the chromosome (the number of codons it will contain).</param>
        public Chromosome(int length)
            : this()
        {
            this.Codons = new object[length];
        }

        /// <summary>
        /// Initialises a new instance of the <see cref="T:Chromosome"/> class.
        /// </summary>
        public Chromosome()
        {
            this.IsOperable = true;
        }

        public double Fitness
        {
            get;
            set;
        }

        public bool IsOperable
        {
            get;
            set;
        }

        public int Length
        {
            get { return this.Codons.Length; }
        }

        public object this[int index]
        {
            get { return this.Codons[index]; }
            set { this.Codons[index] = value; }
        }

        protected internal object[] Codons
        {
            get;
            set;
        }

        public virtual Chromosome Clone(bool copyIsOperable)
        {
            var clone = new Chromosome();
            CloneBase(clone, copyIsOperable);
            return clone;
        }

        protected void CloneBase(Chromosome clone, bool copyIsOperable)
        {
            clone.Codons = (object[])this.Codons.Clone();
            clone.Fitness = this.Fitness;
            if (copyIsOperable)
                clone.IsOperable = this.IsOperable;
        }
    }
}
