﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DarwinDotNet
{
    /// <summary>
    /// Represents a population of chromosomes that can evolve over generations.
    /// </summary>
    /// <remarks>
    /// This class represents a mutable population of chromosomes. It is fundamental in order to simulate the
    /// evolution of a set of individuals (chromosomes).
    /// </remarks>
    public class Population
    {
        /// <summary>
        /// Initialises a new instance of the <see cref="T:Population"/> class, specifying the list of initial
        /// chromosomes.
        /// </summary>
        /// <param name="initialChromosomes">A list of chromosomes of which the population will initially 
        /// consist.</param>
        public Population(List<Chromosome> initialChromosomes)
            : this(initialChromosomes.Count)
        {
            this.Chromosomes = initialChromosomes;
        }

        /// <summary>
        /// Initialises a new instance of the <see cref="T:Population"/> class, specifying the population size.
        /// </summary>
        /// <param name="size">The initial size of the population (the initial number of chromosomes of which
        /// it will consist).</param>
        public Population(int size)
        {
            this.Size = size;
            this.GenerationNumber = 0;
        }

        /// <summary>
        /// Gets or sets the list of generic operators applied to the population to create the next generation.
        /// </summary>
        /// <value>An <see cref="T:IList{Operator}" /> that contains the genetic operators.</value>
        /// <remarks>
        /// Selectors (operators that implement <see cref="T:Selector" />), if contained with the list, should
        /// typically precede operators of other types in standard usage.
        /// </remarks>
        public IList<Operator> GeneticOperators
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the number representing the current generation of the population.
        /// </summary>
        /// <value>The number of the current generation, where zero represents the first generation.</value>
        public int GenerationNumber
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the current size of the population.
        /// </summary>
        /// <value>The size of the population (the number of chromosomes of which it consists).</value>
        /// <remarks>
        /// The value of this property does not necessarily equal the value of the 
        /// <see cref="Chromosomes" />.<see cref="P:System.Collections.Generic.List{Chromosome}.Count" /> 
        /// property at all times. It specifies the target size of the population rather than the actual number 
        /// of chromosomes. The value can be adjusted between generations to simulate a dynamic (growing or 
        /// shrinking) population.
        /// </remarks>
        public int Size
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the list of chromosomes that are current members of the population.
        /// </summary>
        /// <value>The list of chromosomes.</value>
        /// <remarks>
        /// This property does not reference the same object between generations. It will therefore always 
        /// represents a complete list of the chromosomes in the current population at any time.
        /// </remarks>
        public List<Chromosome> Chromosomes
        {
            get;
            set;
        }

        /// <summary>
        /// Gets the list of chromosomes that will compose the next generation of the population.
        /// </summary>
        /// <value>The list of chromsomes.</value>
        /// <remarks>
        /// This property should only be used by the <see cref="T:Selector" /> class to get a reference to the
        /// list of chromosomes in the next generation.
        /// </remarks>
        protected internal List<Chromosome> NewChromosomes
        {
            get;
            private set;
        }

        /// <summary>
        /// Generates random chromosomes until the value of 
        /// <see cref="Chromosomes" />.<see cref="P:System.Collections.Generic.List{Chromosome}.Count" /> equals the
        /// value of <see cref="Size" />.
        /// </summary>
        /// <param name="generator">The function that returns a new random chromosome on each call.</param>
        public void Generate(Func<Population, Chromosome> generator)
        {
            this.Chromosomes = new List<Chromosome>();
            for (int i = this.Chromosomes.Count; i < this.Size; i++)
                this.Chromosomes.Add(generator(this));
        }

#pragma warning disable 1573

        /// <inheritdoc cref="Evolve(Action{Population}, Func{Population, Boolean})" />
        /// <summary>
        /// Evolves the population of chromosomes until the fitness of any member equals or exceeds the specified
        /// target fitness.
        /// </summary>
        /// <param name="targetFitness">The fitness value that, when any member chromosome of the current 
        /// generation of the population equals or exceeds it, terminates evolution.</param>
        /// <remarks>
        /// <inheritdoc select="/para[position() &lt;= last()]" />
        /// </remarks>
        public void Evolve(Action<Population> runSimulation, double targetFitness)
        {
            Evolve(runSimulation, pop => pop.Chromosomes.Any(c => c.Fitness >= targetFitness));
        }

        /// <inheritdoc cref="Evolve(Action{Population}, Func{Population, Boolean})" />
        /// <summary>
        /// Evolves the population of chromosomes until the specified number of generations have been created.
        /// </summary>
        /// <param name="generations">The number of generations for which evolution should proceed before
        /// terminating.</param>
        /// <remarks>
        /// <inheritdoc select="/para[position() &lt;= last()]" />
        /// </remarks>
        public void Evolve(Action<Population> runSimulation, int generations)
        {
            Evolve(runSimulation, pop => pop.GenerationNumber == generations);
        }

        /// <inheritdoc cref="Evolve(Action{Population}, Func{Population, Boolean})" />
        /// <summary>
        /// Evolves the population of chromosomes until the specified duration has elapsed.
        /// </summary>
        /// <param name="duration">The duration for which evolution should proceed before terminating.</param>
        /// <remarks>
        /// <inheritdoc select="/para[position() &lt;= last()]" />
        /// <para>
        /// Due to the previously mentioned reasons, the actual time for which evolution proceeds will be at
        /// least the time specified by <paramref name="duration"/>, though it will never exceed this time by
        /// more than the time required to create the next generation.
        /// </para>
        /// </remarks>
        public void Evolve(Action<Population> runSimulation, TimeSpan duration)
        {
            var endTime = DateTime.Now + duration;
            Evolve(runSimulation, pop => endTime >= DateTime.Now);
        }

#pragma warning restore 1573

        /// <summary>
        /// Evolves the population of chromosomes until the condition specified by the
        /// <paramref name="decideTerminate"/> function is met.
        /// </summary>
        /// <param name="runSimulation">The simulation that is to be run before each new 
        /// generation is created. Optionally evaluates the fitness of each chromosome.</param>
        /// <param name="decideTerminate">The function that returns <see langword="true"/> when evolution should
        /// terminate, and <see langword="false"/> until then.</param>
        /// <exception cref="InvalidOperationException">A parameter of the <see cref="T:Population" /> is invalid
        /// or has not been specified.</exception>
        /// <remarks>
        /// <para>
        /// This method synchronously evolves the population of chromosomes and only returns once the condition 
        /// specified by the <paramref name="decideTerminate"/> parameter is met. The 
        /// <paramref name="decideTerminate"/> function is evaluated once per generation after the function
        /// <paramref name="runSimulation"/> has been called. For this reason, it is suggested that halting
        /// evolution at an arbitrary time (for example, based on a user input event) should be done using
        /// standard threading mechanisms. A reference to the most recent generation of chromosomes is always
        /// available via the <see cref="Chromosomes"/> property, meaning that the simulation can be terminated 
        /// at any time and a complete list of chromosomes in the most recent generation will assuredly be 
        /// available.
        /// </para>
        /// </remarks>
        public void Evolve(Action<Population> runSimulation, Func<Population, bool> decideTerminate)
        {
            if (this.Chromosomes == null)
                throw new InvalidOperationException("The list of chromosomes is null.");
            if (this.GeneticOperators == null)
                throw new InvalidOperationException("The list of genetic operators to perform is null.");
            //if (!this.GeneticOperators.Any(op => op is Selector))
            //    throw new InvalidOperationException("The list of genetic operators must contain a selector.");

            while (true)
            {
                // Run simulation on current population.
                if (runSimulation != null)
                    runSimulation(this);

                // Decide whether to terminate evolution now.
                if (decideTerminate(this))
                    break;

                // Perform each genetic operator.
                Operator curOperator;

                this.NewChromosomes = new List<Chromosome>(this.Size);
                for (int i = 0; i < this.GeneticOperators.Count; i++)
                {
                    curOperator = this.GeneticOperators[i];
                    if (!(curOperator is Selector))
                    {
                        if (this.NewChromosomes.Count > 0)
                            this.Chromosomes = this.NewChromosomes;
                    }
                    curOperator.Perform(this);
                }

                // Increment generation number.
                this.GenerationNumber++;
            }
        }
    }
}
