﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using DarwinDotNet;
using DarwinDotNet.Expressions;
using DarwinDotNet.Selectors;

namespace SymbolicRegression
{
    internal static class Program
    {
        internal static readonly double _maxFitness = 500000;
        internal static readonly Func<double, double, double, double> _testFunction =
            (x, y, z) => x * x * z - x * y * z + x * y * y - z;
        internal static TestCase[] _testCases;

        internal static Random _rand = new Random();

        internal static void Main(string[] args)
        {
            TestCase curCase;
            _testCases = new TestCase[30];
            for (int i = 0; i < _testCases.Length; i++)
            {
                curCase = new TestCase()
                {
                    X = _rand.NextDouble() * 100,
                    Y = _rand.NextDouble() * 100,
                    Z = _rand.NextDouble() * 100
                };
                curCase.ExpectedResult = _testFunction(curCase.X, curCase.Y, curCase.Z);
                _testCases[i] = curCase;
            }

            var population = new Population(2000);
            population.GeneticOperators = new Operator[] { 
                new ElitistSelector() { ElitistCountFunction = (p => p.GenerationNumber < 80 ? 500 : 250) },
                new FitnessProportionalSelector(),
                new SimpleMutator() { ProbabilityFunction = (p => 0.14d) },
                new SimpleTranspositor() { ProbabilityFunction = (p => 0.25d) },
                new SimpleRecombinator() { ProbabilityFunction = (p => p.GenerationNumber < 120 ? 0.92 : 0.0)},
            };
            population.Generate(GenerateRandomChromosome);
            population.Evolve(RunSimulation, _maxFitness - 0.5d);

            population.Chromosomes.Sort(new ChromosomeFitnessComparer());
            var bestChromosome = population.Chromosomes.Last();
            Console.WriteLine("Found solution in generation {0:#0}.", population.GenerationNumber);
            Console.WriteLine("Best chromosome in final generation: {0} (fitness {1:#0})", bestChromosome,
                bestChromosome.Fitness);
#if DEBUG
            Console.ReadKey(true);
#endif
        }

        public static void RunSimulation(Population population)
        {
            var chromosomes = population.Chromosomes;
            Func<Expression<MathSymbol>, MathSymbol, Expression<MathSymbol>>
                generateExpression = (parent, symbol) => new MathExpression(symbol, parent);

            //for (int i = 0; i < chromosomes.Count; i++)
            Parallel.For(0, chromosomes.Count, delegate(int i)
            {
                var chromosome = (ExpressionChromosome<MathSymbol>)chromosomes[i];
                var expression = (MathExpression)chromosome.GenerateExpressionTree(generateExpression);
                TestCase curCase;
                double sumResidualsSquared = 0;

                for (int j = 0; j < _testCases.Length; j++)
                {
                    curCase = _testCases[j];
                    sumResidualsSquared += (expression.Evaluate(curCase.X, curCase.Y, curCase.Z) -
                        curCase.ExpectedResult).Square();
                }

                if (!double.IsNaN(sumResidualsSquared))
                {
                    chromosome.Fitness = Math.Max(_maxFitness - Math.Sqrt(sumResidualsSquared /
                        _testCases.Length), 0);
                }
            });

            Console.WriteLine("Generation #{0:#00000} | Best Fitness: {1:#00000} | Average Fitness: {2:#000000}",
                population.GenerationNumber, population.Chromosomes.Max(c => c.Fitness),
                population.Chromosomes.Average(c => c.Fitness));
        }

        public static Chromosome GenerateRandomChromosome(Population population)
        {
            var chromosome = new ExpressionChromosome<MathSymbol>(10, 11);
            for (int i = 0; i < chromosome.HeadLength; i++)
                chromosome[i] = (MathSymbol)_rand.Next(0, (int)MathSymbol.Max);
            for (int i = chromosome.HeadLength; i < chromosome.Length; i++)
                chromosome[i] = (MathSymbol)_rand.Next((int)MathSymbol.FunctionMax, (int)MathSymbol.Max);

            return chromosome;
        }
    }

    public struct TestCase
    {
        public double ExpectedResult;
        public double X, Y, Z;
    }
}
